package javafold;

import java.util.Hashtable;
import java.util.Map;
import javax.swing.text.Segment;
import org.gjt.sp.jedit.Buffer;
import org.gjt.sp.jedit.buffer.JEditBuffer;
import org.gjt.sp.jedit.buffer.FoldHandler;

public class JavaFold extends FoldHandler {

	/** folding based on opened/closed braces */
	private static final byte BRACE = 0;
	/** fold multi line comment */
	private static final byte SINGLE_LINE_COMMENT = (byte) 1;
	/** fold consecutive single line coments */
	private static final byte MULTI_LINE_COMMENT = (byte) 2;
	/** fold consecutive lines ending with ';' */
	private static final byte GROUP = (byte) 3;

	/** contains Buffer, LineData pairs */
	// LineData contains informations for the last processed line
	// it would be a little faster if currently edited line was also remembered
	private static Hashtable buffers = new Hashtable();

	private static class LineData {
		LineData previous;
		int index;
		short openedBraces; // int would be safer, but who opens 30000 braces and doesn't close them?
		byte type;

		LineData(LineData previous, int index, short openedBraces, byte type) {
			this.previous = previous;
			this.index = index;
			this.openedBraces = openedBraces;
			this.type = type;
		}
	}

	public JavaFold() {
		super("java");
	}

	static Map getBuffers() {
		return buffers;
	}

	public int getFoldLevel(JEditBuffer buffer, int lineIndex, Segment segment) {
		buffer.getLineText(lineIndex, segment);

		synchronized (buffer) {
			LineData lineData = (LineData) buffers.get(buffer);

			while (true) {
				if (lineData == null || lineData.index < lineIndex) {
					break;
				}

				lineData = lineData.previous;
			}

			short bracesBefore;
			int typeBefore;

			if (lineData == null) {
				bracesBefore = 0;
				typeBefore = BRACE;
			} else {
				bracesBefore = lineData.openedBraces;
				typeBefore = lineData.type;
			}

			short braces = bracesBefore;
			byte type = (typeBefore == MULTI_LINE_COMMENT ? MULTI_LINE_COMMENT : BRACE);

			short lineOpenedBraces = 0;
			boolean braceClosed = false;
			char stringChar = 0;
			char lastChar = 0;
			boolean outsideMultiLineComment = false;
			boolean outsideClosingBrace = false; // still false if only comment is outside
			boolean notSlash = false; // true if there is a text before single line comment
			boolean groupChar = false; // true if the last char [before comment] was ';'

			char c;
			int z = segment.offset + segment.count;
			for (int i = segment.offset; i < z; i++, lastChar = c) {
				c = segment.array[i];

				if (c <= ' ') {
					continue;
				}

				if (c != '/') {
					notSlash = true;
				}

				if (type == MULTI_LINE_COMMENT) {
					if (c == '/' && lastChar == '*') {
						type = BRACE;
						c = 0; //  /* *// won't be detected as a single line comment
					}
				} else if (stringChar != 0) {
					if (lastChar == '\\') {
						c = 0;
					} else if (c == stringChar) {
						stringChar = 0;
					}
				} else {
					outsideMultiLineComment = true;

					if (c == '/') {
						if (lastChar == '/') {
							if (!notSlash) {
								type = SINGLE_LINE_COMMENT;
							}

							break;
						}

						// i should set outsideClosingBrace = true and after following '/' or '*'
						// set it to current value, but i don't wanna
						// now last line will be hidden in example below:
						// {
						// } /
					} else {
						if (c == '*' && lastChar == '/') {
							type = MULTI_LINE_COMMENT;
							c = 0; //  /*/ won't be detected as a closed comment

							continue;
						}

						groupChar = (c == ';');
						outsideClosingBrace = true;

						if (c == '"' || c == '\'') {
							stringChar = c;
						} else if (c == '{') {
							braces++;
							lineOpenedBraces++;
						} else if (c == '}') {
							braceClosed = true;

							if (lineOpenedBraces > 0) {
								lineOpenedBraces--;
							} else {
								outsideClosingBrace = false;
							}

							if (braces > 0) {
								braces--;
							}
						}
					}
				}
			}

			// more precisely : type == BRACE - > type != SINGLE_LINE_COMMENT && type != MULTI_LINE_COMMENT
			if (type == BRACE && typeBefore != MULTI_LINE_COMMENT &&
					lineOpenedBraces == 0 && !braceClosed && groupChar) {

				type = GROUP;
			}

			int level = braces - lineOpenedBraces;

			// brace that closes the code block should be hidden
			// only if no uncommented code is following it
			if (type == BRACE && braceClosed && !outsideClosingBrace) {
				level++;
			}

			if ((typeBefore == MULTI_LINE_COMMENT && !outsideMultiLineComment) ||
					(typeBefore == SINGLE_LINE_COMMENT && type == SINGLE_LINE_COMMENT) ||
					(typeBefore == GROUP && type == GROUP)) {

				level++;
			}

			if (braces != bracesBefore || type != typeBefore) {
				lineData = new LineData(lineData, lineIndex, braces, type);
			}

			// maybe it would be a little faster to
			// do this only if lineData changes
			if (lineData == null) {
				buffers.remove(buffer);
			} else {
				buffers.put(buffer, lineData);
			}

			return level;
		}
	}

	// this method assumes that buffers map contains informations for
	// last line. this may not be true. adding last line to buffers map
	// would correct this problem, but buffers doesn't contain
	// necessary informations only while calculating folds, so remembered
	// informations may be already invalid
	// maybe i should wait here until getFoldLevel() is called for the last line
	boolean isCommentFold(Buffer buffer, int lineIndex) {
		synchronized (buffer) {
			LineData lineData = (LineData) buffers.get(buffer);

			while (true) {
				if (lineData == null || lineData.index <= lineIndex) {
					break;
				}

				lineData = lineData.previous;
			}

			if (lineData == null) {
				return false;
			}

			return lineData.type == SINGLE_LINE_COMMENT || lineData.type == MULTI_LINE_COMMENT;
		}
	}
}
